unit IPLocationUnit;

interface

uses
  System.Classes, System.SysUtils, winapi.WinSock, System.Math,
  System.SyncObjs;

type
  TIPLocation = class(TObject)
  private
    QQWryFileName: string;
    QQWryFileStream: TBufferedFileStream;
    QQWryFileSize: Cardinal;
    IPRecordNum: Cardinal;
    FirstIPIndexOffset, LastIPIndexOffset: Cardinal;
    FLock: TCriticalSection;
    function GetIPLocation(IPLocationOffset: Cardinal): TStringlist;
    function GetIPMsg(IPRecordID: Cardinal): TStringlist;
    function GetIPRecordID(IP: string): Cardinal;
    function GetIPValue(IP: string): Cardinal;
  public
    function GetQQWryFileName: string;
    function GetQQWryFileSize: Cardinal;
    function GetIPRecordNum: Cardinal;
    function GetQQWryDate: TDate;
    function GetQQWryDataFrom: string;
    constructor Create(cQQWryFileName: string);
    destructor Destroy; override;
    function GetLocation(IP: string): string;
  end;

implementation

var
  __IPLocation: TIPLocation;

function IPLocation: TIPLocation;
begin
  if __IPLocation = nil then
    __IPLocation := TIPLocation.Create(ExtractFilePath(ParamStr(0)) +
      'qqwry.dat');
  Result := __IPLocation;
end;

{ TIPLocation }

constructor TIPLocation.Create(cQQWryFileName: string);
begin
  inherited Create;
  FLock := TCriticalSection.Create;
  QQWryFileName := cQQWryFileName;
  QQWryFileStream := TBufferedFileStream.Create(QQWryFileName,
    fmOpenRead or fmShareDenyWrite, 0);
  QQWryFileSize := QQWryFileStream.Size;
  QQWryFileStream.Read(FirstIPIndexOffset, 4);
  QQWryFileStream.Read(LastIPIndexOffset, 4);
  IPRecordNum := (LastIPIndexOffset - FirstIPIndexOffset) div 7 + 1;
end;

destructor TIPLocation.Destroy;
begin

  QQWryFileStream.Free;
  FLock.Free;
  inherited Destroy;
end;

function TIPLocation.GetIPLocation(IPLocationOffset: Cardinal): TStringlist;
const
  // ʵϢִλõضģʽ
  REDIRECT_MODE_1 = 1;
  REDIRECT_MODE_2 = 2;
var
  RedirectMode: byte;
  CountryFirstOffset, CountrySecondOffset: Cardinal;
  CountryMsg, AreaMsg: AnsiString;
  //

  function ReadString(StringOffset: Cardinal): AnsiString;
  var
    ReadByte: AnsiChar;
  begin
    Result := '';
    QQWryFileStream.Seek(StringOffset, soFromBeginning);
    QQWryFileStream.Read(ReadByte, 1);
    while ord(ReadByte) <> 0 do
    begin
      Result := Result + ReadByte;
      QQWryFileStream.Read(ReadByte, 1);
    end;
  end;
//

  function ReadArea(AreaOffset: Cardinal): AnsiString;
  var
    ModeByte: byte;
    ReadAreaOffset: Cardinal;
  begin
    ReadAreaOffset := 0;
    QQWryFileStream.Seek(AreaOffset, soFromBeginning);
    QQWryFileStream.Read(ModeByte, 1);
    if (ModeByte = REDIRECT_MODE_1) or (ModeByte = REDIRECT_MODE_2) then
    begin
      QQWryFileStream.Read(ReadAreaOffset, 3);
      if ReadAreaOffset = 0 then
        Result := 'δ֪'
      else
        Result := ReadString(ReadAreaOffset);
    end
    else
    begin
      Result := ReadString(AreaOffset);
    end;
  end;

begin
  CountryFirstOffset := 0;
  CountrySecondOffset := 0;
  // 4ֽڣ4ֽΪIPϢIPַеֹIPֵ
  QQWryFileStream.Seek(IPLocationOffset + 4, soFromBeginning);
  // ȡϢضģʽֵ
  QQWryFileStream.Read(RedirectMode, 1);
  // ضģʽ1Ĵ
  if RedirectMode = REDIRECT_MODE_1 then
  begin
    // ģʽֵΪ13ֽڵΪϢضƫֵ
    QQWryFileStream.ReadData(CountryFirstOffset, 3);
    // ض
    QQWryFileStream.Seek(CountryFirstOffset, soFromBeginning);
    // ڶζȡϢضģʽ
    QQWryFileStream.Read(RedirectMode, 1);
    // ڶضģʽΪģʽ2Ĵ
    if RedirectMode = REDIRECT_MODE_2 then
    begin
      // 3ֽڵݼΪڶضƫֵ
      QQWryFileStream.ReadData(CountrySecondOffset, 3);
      // ȡڶضƫֵµֵַΪϢ
      CountryMsg := ReadString(CountrySecondOffset);
      // һضģʽΪ1ضȡĵڶضģʽΪ2
      // ϢڵһιϢƫֵĺ
      QQWryFileStream.Seek(CountryFirstOffset + 4, soFromBeginning);
      // ڶضģʽģʽ2Ĵ
    end
    else
    begin
      CountryMsg := ReadString(CountryFirstOffset);
    end;
    // ضģʽ1¶Ϣֵ
    AreaMsg := ReadArea(QQWryFileStream.Position);
    // ضģʽ2Ĵ
  end
  else if RedirectMode = REDIRECT_MODE_2 then
  begin
    QQWryFileStream.ReadData(CountrySecondOffset, 3);
    CountryMsg := ReadString(CountrySecondOffset);
    AreaMsg := ReadArea(IPLocationOffset + 8);
    // ضģʽĴŵļIPַϢ
  end
  else
  begin
    CountryMsg := ReadString(QQWryFileStream.Position - 1);
    AreaMsg := ReadArea(QQWryFileStream.Position);
  end;
  Result := TStringlist.Create;
  Result.Add(string(CountryMsg));
  Result.Add(string(AreaMsg));
end;

function TIPLocation.GetIPMsg(IPRecordID: Cardinal): TStringlist;
var
  aryStartIP: array [1 .. 4] of byte;
  strStartIP: string;
  EndIPOffset: Cardinal;
  aryEndIP: array [1 .. 4] of byte;
  strEndIP: string;
  i: integer;
begin
  EndIPOffset := 0;

  // ݼ¼IDƵü¼ŵ
  QQWryFileStream.Seek(FirstIPIndexOffset + (IPRecordID - 1) * 7,
    soFromBeginning);
  // ǰ4ֽΪʼIPַ
  QQWryFileStream.Read(aryStartIP, 4);
  // 3ֽƫֵ
  // QQWryFileStream.Read(EndIPOffset, 3);
  QQWryFileStream.ReadData(EndIPOffset, 3);
  // 
  QQWryFileStream.Seek(EndIPOffset, soFromBeginning);
  // ǰ4ֽΪֹIPַ
  QQWryFileStream.Read(aryEndIP, 4);

  // ֹIPַתΪֵʽ
  strStartIP := '';
  for i := 4 downto 1 do
  begin
    if i <> 1 then
      strStartIP := strStartIP + IntToStr(aryStartIP[i]) + '.'
    else
      strStartIP := strStartIP + IntToStr(aryStartIP[i]);
  end;
  strEndIP := '';
  for i := 4 downto 1 do
  begin
    if i <> 1 then
      strEndIP := strEndIP + IntToStr(aryEndIP[i]) + '.'
    else
      strEndIP := strEndIP + IntToStr(aryEndIP[i]);
  end;
  Result := TStringlist.Create;
  Result.Add(strStartIP);
  Result.Add(strEndIP);
  // ȡ¼µIPַϢ
  // ͳһģƫֵ ֹIPַĴλ ۹ϢֹIPַλú
  Result.AddStrings(GetIPLocation(EndIPOffset));
end;

function TIPLocation.GetIPRecordID(IP: string): Cardinal;

  function SearchIPRecordID(IPRecordFrom, IPRecordTo, IPValue: Cardinal)
    : Cardinal;
  var
    CompareIPValue1, CompareIPValue2: Cardinal;
  begin
    Result := 0;
    CompareIPValue1 := 0;
    CompareIPValue2 := 0;
    QQWryFileStream.Seek(FirstIPIndexOffset + ((IPRecordTo - IPRecordFrom) div 2
      + IPRecordFrom - 1) * 7, soFromBeginning);
    QQWryFileStream.Read(CompareIPValue1, 4);
    QQWryFileStream.Seek(FirstIPIndexOffset + ((IPRecordTo - IPRecordFrom) div 2
      + IPRecordFrom) * 7, soFromBeginning);
    QQWryFileStream.Read(CompareIPValue2, 4);
    // ҵ
    if (IPValue >= CompareIPValue1) and (IPValue < CompareIPValue2) then
    begin
      Result := (IPRecordTo - IPRecordFrom) div 2 + IPRecordFrom;
    end
    else
      // 
      if IPValue > CompareIPValue1 then
      begin
        Result := SearchIPRecordID((IPRecordTo - IPRecordFrom) div 2 +
          IPRecordFrom + 1, IPRecordTo, IPValue);
      end
      else
        // ǰ
        if IPValue < CompareIPValue1 then
        begin
          Result := SearchIPRecordID(IPRecordFrom, (IPRecordTo - IPRecordFrom)
            div 2 + IPRecordFrom - 1, IPValue);
        end;
  end;

begin
  Result := SearchIPRecordID(1, GetIPRecordNum, GetIPValue(IP));
end;

function TIPLocation.GetIPRecordNum: Cardinal;
begin
  Result := IPRecordNum;
end;

function TIPLocation.GetIPValue(IP: string): Cardinal;
var
  tsIP: TStringlist;
  i: integer;

  function SplitStringToStringlist(aString: string; aSplitChar: string)
    : TStringlist;
  begin
    Result := TStringlist.Create;
    while pos(aSplitChar, aString) > 0 do
    begin
      Result.Add(copy(aString, 1, pos(aSplitChar, aString) - 1));
      aString := copy(aString, pos(aSplitChar, aString) + 1,
        length(aString) - pos(aSplitChar, aString));
    end;
    Result.Add(aString);
  end;

begin
  tsIP := SplitStringToStringlist(IP, '.');
  Result := 0;
  for i := 3 downto 0 do
  begin
    Result := Result + StrToInt(tsIP[i]) * trunc(power(256, 3 - i));
  end;
end;

function TIPLocation.GetLocation(IP: string): string;
begin
  FLock.Enter;
  try
    Result := GetIPMsg(GetIPRecordID(IP))[2] + GetIPMsg(GetIPRecordID(IP))[3];
  finally
    FLock.Leave;
  end;
end;

function TIPLocation.GetQQWryDataFrom: string;
begin
  Result := GetIPMsg(GetIPRecordNum)[2];
end;

function TIPLocation.GetQQWryDate: TDate;
var
  DateString: string;
begin
  DateString := GetIPMsg(GetIPRecordNum)[3];
  DateString := copy(DateString, 1, pos('IP', DateString) - 1);
  DateString := StringReplace(DateString, '', '-',
    [rfReplaceAll, rfIgnoreCase]);
  DateString := StringReplace(DateString, '', '-',
    [rfReplaceAll, rfIgnoreCase]);
  DateString := StringReplace(DateString, '', '-',
    [rfReplaceAll, rfIgnoreCase]);
  Result := StrToDate(DateString);
end;

function TIPLocation.GetQQWryFileName: string;
begin
  Result := QQWryFileName;
end;

function TIPLocation.GetQQWryFileSize: Cardinal;
begin
  Result := QQWryFileSize;
end;

initialization

finalization

if __IPLocation <> nil then
  __IPLocation.Free;

end.
